package DMA;

import ESABus :: *;
import GetPut :: *;
import Vector :: *;
import FIFO   :: *;

interface DMA_StreamToMemory;
    interface ESABusWr_Master_Fabric m_wr;
    interface ESABusWr_Slave_Fabric s_wr;
    interface ESABusRd_Slave_Fabric s_rd;
    method Bool doneInterrupt();

    interface Put#(Bit#(8)) data;
endinterface

module mkDMA_StreamToMemory(DMA_StreamToMemory);

    ESABusWr_Master master <- mkESABusWr_Master();

    ESABusWr_Slave slave_write <- mkESABusWr_Slave();
    ESABusRd_Slave slave_read <- mkESABusRd_Slave();

    Vector#(2, Reg#(Bit#(8))) addr <- replicateM(mkRegU);
    Reg#(Bool) idle <- mkReg(True);
    Vector#(2, Reg#(Bit#(8))) transfer_length <- replicateM(mkRegU);

    FIFO#(Tuple2#(Bit#(16), Bit#(16))) requests <- mkFIFO();

    // Address Map:
    // 0: Write Address Low
    // 1: Write Address High
    // 2: Transfer Length Low
    // 3: Transfer Length High
    // 4: Start
    // 5: Done
    rule handleSlaveWrite;
        let r <- slave_write.request.get();
        // We escape the address as we only care for the last three bits
        if((tpl_1(r) & 'h7) == 0) addr[0] <= tpl_2(r);
        else if((tpl_1(r) & 'h7) == 1) addr[1] <= tpl_2(r);
        else if((tpl_1(r) & 'h7) == 2) transfer_length[0] <= tpl_2(r);
        else if((tpl_1(r) & 'h7) == 3) transfer_length[1] <= tpl_2(r);
        else if((tpl_1(r) & 'h7) == 4) begin
            Bit#(16) a_tmp = pack(readVReg(addr)); // Transform Vector#(a, Reg#(b)) to Vector#(a, b) to Bit#(a * b_sz)
            Bit#(16) b_tmp = pack(readVReg(transfer_length));
            requests.enq(tuple2(a_tmp, b_tmp));
        end
    endrule

    rule handleSlaveRead;
        let r <- slave_read.request.get();
        // Response 0 if unused register is read.
        // Only register 4 is read.
        if((r & 'h7) == 4) slave_read.response.put(extend(pack(idle)));
        else slave_read.response.put(0);
    endrule

    Reg#(Bit#(16)) currentTransfer <- mkReg(0);
    Reg#(Bit#(16)) currentAddress <- mkReg(0);

    rule startTransfer if(idle);
        currentTransfer <= tpl_2(requests.first());
        currentAddress <= tpl_1(requests.first());
        requests.deq();
        idle <= False;
    endrule

    Wire#(Bool) done <- mkDWire(False);
    FIFO#(Bit#(8)) dataIn <- mkFIFO();

    rule process if(!idle);
        let currentTransferT = currentTransfer - 1;
        if(currentTransferT == 0) begin
            idle <= True;
            done <= True;
        end
        currentTransfer <= currentTransferT;
        currentAddress <= currentAddress + 1;
        master.request.put(tuple2(currentAddress, dataIn.first()));
        dataIn.deq();
    endrule

    interface m_wr = master.fab;
    interface s_wr = slave_write.fab;
    interface s_rd = slave_read.fab;

    interface data = toPut(dataIn);
    interface doneInterrupt = done;
endmodule

interface DMA_MemoryToStream;
    interface ESABusRd_Master_Fabric m_rd;
    interface ESABusWr_Slave_Fabric s_wr;
    interface ESABusRd_Slave_Fabric s_rd;
    method Bool doneInterrupt();

    interface Get#(Bit#(8)) data;
endinterface

module mkDMA_MemoryToStream(DMA_MemoryToStream);

    ESABusRd_Master master <- mkESABusRd_Master();

    ESABusWr_Slave slave_write <- mkESABusWr_Slave();
    ESABusRd_Slave slave_read <- mkESABusRd_Slave();

    Vector#(2, Reg#(Bit#(8))) addr <- replicateM(mkRegU);
    Reg#(Bool) idle <- mkReg(True);
    Vector#(2, Reg#(Bit#(8))) transfer_length <- replicateM(mkRegU);

    FIFO#(Tuple2#(Bit#(16), Bit#(16))) requests <- mkFIFO();

    // Address Map:
    // 0: Write Address Low
    // 1: Write Address High
    // 2: Transfer Length Low
    // 3: Transfer Length High
    // 4: Start
    // 5: Done
    rule handleSlaveWrite;
        let r <- slave_write.request.get();
        // We escape the address as we only care for the last three bits
        if((tpl_1(r) & 'h7) == 0) addr[0] <= tpl_2(r);
        else if((tpl_1(r) & 'h7) == 1) addr[1] <= tpl_2(r);
        else if((tpl_1(r) & 'h7) == 2) transfer_length[0] <= tpl_2(r);
        else if((tpl_1(r) & 'h7) == 3) transfer_length[1] <= tpl_2(r);
        else if((tpl_1(r) & 'h7) == 4) begin
            Bit#(16) a_tmp = pack(readVReg(addr)); // Transform Vector#(a, Reg#(b)) to Vector#(a, b) to Bit#(a * b_sz)
            Bit#(16) b_tmp = pack(readVReg(transfer_length));
            requests.enq(tuple2(a_tmp, b_tmp));
        end
    endrule

    rule handleSlaveRead;
        let r <- slave_read.request.get();
        // Response 0 if unused register is read.
        // Only register 4 is read.
        if((r & 'h7) == 4) slave_read.response.put(extend(pack(idle)));
        else slave_read.response.put(0);
    endrule

    Reg#(Bit#(16)) currentTransfer <- mkReg(0);
    Reg#(Bit#(16)) currentAddress <- mkReg(0);

    rule startTransfer if(idle);
        currentTransfer <= tpl_2(requests.first());
        currentAddress <= tpl_1(requests.first());
        requests.deq();
        idle <= False;
    endrule

    Wire#(Bool) done <- mkDWire(False);

    rule process if(!idle);
        let currentTransferT = currentTransfer - 1;
        if(currentTransferT == 0) begin
            idle <= True;
            done <= True;
        end
        currentTransfer <= currentTransferT;
        currentAddress <= currentAddress + 1;
        master.request.put(currentAddress);
    endrule

    interface m_rd = master.fab;
    interface s_wr = slave_write.fab;
    interface s_rd = slave_read.fab;

    interface data = master.response;
    interface doneInterrupt = done;
endmodule

endpackage