package MedianFilter;

    import Types :: *;
    import ClientServer :: *;
    import GetPut :: *;
    import Vector :: *;
    import FIFO :: *;
    import Connectable :: *;

    import BlueCheck :: *;
    import Median :: *;
    import ColorConverter :: *;

    import StmtFSM :: *;

    typedef 20 ImageWidth;
    typedef 20 ImageHeight;
    typedef 3 KernelSize;
    typedef TMul#(ImageHeight, ImageWidth) ImagePixels;
    typedef TAdd#(ImagePixels, TMul#(2, ImageWidth)) ImagePixelsTotal;
    typedef TSub#(ImageWidth, KernelSize) BufferSize;

    module mkMedianFilter(Server#(GrayScale, GrayScale));
        Server#(GrayScale, Vector#(9, GrayScale)) kernel <- mkKernelBuffer();

        Median median <- mkMedian();

        FIFO#(GrayScale) in <- mkFIFO();

        Reg#(UInt#(TLog#(TMul#(2, ImagePixelsTotal)))) inputCounter[2] <- mkCReg(2, 0);

        function UInt#(TLog#(TMul#(2, ImagePixelsTotal))) incrWithMax(UInt#(TLog#(TMul#(2, ImagePixelsTotal))) r);
            let t = r + 1;
            if(t == fromInteger(2 * valueOf(ImagePixelsTotal))) t = 0;
            return t;
        endfunction

        rule inForward if(((inputCounter[0] & 1) == 0) 
                              && inputCounter[0] < fromInteger(2 * valueOf(ImagePixels))
                              );
            kernel.request.put(in.first());
            in.deq();
            inputCounter[0] <= incrWithMax(inputCounter[0]);
        endrule

        rule inFeed if(((inputCounter[0] & 1) == 0)
                           && inputCounter[0] >= fromInteger(2 * valueOf(ImagePixels))
                        );
            kernel.request.put(0);
            inputCounter[0] <= incrWithMax(inputCounter[0]);
        endrule

        rule outDiscard if(((inputCounter[1] & 1) == 1) 
                           && inputCounter[1] < fromInteger(2 * valueOf(ImageWidth))
                           );
            let d <- kernel.response.get();
            inputCounter[1] <= incrWithMax(inputCounter[1]);
        endrule

        rule outForward if(((inputCounter[1] & 1) == 1) 
                               && inputCounter[1] >= fromInteger(2 * valueOf(ImageWidth))
                                );
            let d <- kernel.response.get();
            median.request.put(d);
            inputCounter[1] <= incrWithMax(inputCounter[1]);
        endrule

        interface request = toPut(in);
        interface response = median.response;
    endmodule

    module mkMedianFilterTest(Empty);

        Server#(GrayScale, GrayScale) medianFilter <- mkMedianFilter();

        Reg#(UInt#(32)) cntr <- mkReg(0);
        Reg#(UInt#(32)) cntr2 <- mkReg(0);

        Stmt fsm = {
            par
                $display("Starting test");
                for(cntr <= 0; cntr < fromInteger(2 * valueOf(ImagePixels)); cntr <= cntr + 1) seq
                    action
                        $display("Putting in %d", cntr);
                        medianFilter.request.put(pack(truncate(cntr + 1)));
                    endaction
                endseq
                for(cntr2 <= 0; cntr2 < fromInteger(2 * valueOf(ImagePixels)); cntr2 <= cntr2 + 1) seq
                    action
                        let d <- medianFilter.response.get();
                        $display("Fetching %d in %d", d, cntr2);
                    endaction
                endseq
            endpar
        };

        mkAutoFSM(fsm);
    endmodule

    module mkKernelBuffer(Server#(GrayScale, Vector#(9, GrayScale)));
        Reg#(Bool) valueFetched[2] <- mkCReg(2, False);

        FIFO#(GrayScale) in <- mkFIFO();
        FIFO#(Vector#(9, GrayScale)) out <- mkFIFO();

        Reg#(Vector#(9, GrayScale)) bufferReg <- mkReg(unpack(0));

        FIFO#(Vector#(2, GrayScale)) bufferFIFO <- mkAlwaysFullSizedFIFO();

        Wire#(Vector#(2, GrayScale)) bufferWire <- mkWire();

        rule feedBufferRegular;
            Vector#(9, GrayScale) tmpBuf;

            tmpBuf[0] = in.first(); in.deq();
            for(Integer i = 1; i < 9; i = i + 1) begin
                tmpBuf[i] = bufferReg[i - 1];
            end

            Vector#(2, GrayScale) t = bufferFIFO.first(); bufferFIFO.deq();
            tmpBuf[3] = t[0];
            tmpBuf[6] = t[1];

            out.enq(tmpBuf);
            bufferReg <= tmpBuf;

            t[0] = bufferReg[2];
            t[1] = bufferReg[5];
            bufferWire <= t;

            valueFetched[1] <= False;
        endrule

        rule feedBuffer;
            bufferFIFO.enq(bufferWire);
        endrule

        interface request = toPut(in);
        interface response = toGet(out);
    endmodule

    module mkKernelBufferTest(Empty);

        Server#(GrayScale, Vector#(9, GrayScale)) kernel <- mkKernelBuffer();

        Reg#(UInt#(32)) cntr <- mkReg(0);

        Stmt fsm = {
            seq
                $display("Starting test");
                for(cntr <= 0; cntr < 40; cntr <= cntr + 1) par
                    action
                        let d <- kernel.response.get();
                        $display("Fetching in %d", cntr);
                        $display(fshow(d));
                    endaction
                    action
                        $display("Putting in %d", cntr);
                        kernel.request.put(pack(truncate(cntr + 1)));
                    endaction
                endpar
            endseq
        };

        mkAutoFSM(fsm);
    endmodule

    module mkAlwaysFullSizedFIFO(FIFO#(t))
        provisos(Bits#(t, t_sz));
        Reg#(UInt#(TLog#(BufferSize))) writePntr[3] <- mkCReg(3, 0);
        Reg#(UInt#(TLog#(BufferSize))) readPntr[3] <- mkCReg(3, 0);

        Vector#(BufferSize, Reg#(t)) buffer <- replicateM(mkReg(unpack(0)));

        function UInt#(TLog#(BufferSize)) incrWithMax(Reg#(UInt#(TLog#(BufferSize))) r);
            let t = r + 1;
            if(t == fromInteger(valueOf(BufferSize))) t = 0;
            return t;
        endfunction

        method Action enq (t x) if(incrWithMax(writePntr[1]) == readPntr[1]);
            writePntr[1] <= incrWithMax(writePntr[1]);
            buffer[writePntr[1]] <= x;
        endmethod

        method Action deq if(readPntr[0] == writePntr[0]);
            readPntr[0] <= incrWithMax(readPntr[0]);
        endmethod

        interface first = buffer[readPntr[0]];

        method Action clear();
            writePntr[2] <= 0;
            readPntr[2] <= 1;
        endmethod
    endmodule

    module mkAlwaysFullSizedFIFOTest(Empty);

        FIFO#(UInt#(32)) f <- mkAlwaysFullSizedFIFO();

        Reg#(UInt#(32)) cntr <- mkReg(0);

        Stmt fsm = {
            seq
                $display("Starting test");
                for(cntr <= 0; cntr < 40; cntr <= cntr + 1) par
                    action
                        $display("Fetching %d in %d", f.first(), cntr);
                        f.deq();
                    endaction
                    action
                        $display("Putting in %d", cntr);
                        f.enq(cntr + 1);
                    endaction
                endpar
            endseq
        };

        mkAutoFSM(fsm);
    endmodule


    module mkRGBToMedian(Server#(RGB, GrayScale));
        Server#(RGB, GrayScale) gray <- mkGray();
        Server#(GrayScale, GrayScale) medianFilter <- mkMedianFilter();
        mkConnection(gray.response, medianFilter.request);

        interface request = gray.request;
        interface response = medianFilter.response;
    endmodule
endpackage